@@ -300,8 +300,8 @@ match_prefix (const char *path, const char *prefix)
 }
 
 static gboolean
-g_daemon_file_contains_file (GFile *parent,
-			     GFile *descendant)
+g_daemon_file_prefix_matches (GFile *parent,
+			      GFile *descendant)
 {
   GDaemonFile *parent_daemon = G_DAEMON_FILE (parent);
   GDaemonFile *descendant_daemon = G_DAEMON_FILE (descendant);
@@ -629,7 +629,6 @@ do_async_path_call (GFile *file,
 {
   GDaemonFile *daemon_file = G_DAEMON_FILE (file);
   va_list var_args;
-  GError *error;
   AsyncPathCall *data;
 
   data = g_new0 (AsyncPathCall, 1);
@@ -646,8 +645,6 @@ do_async_path_call (GFile *file,
   data->callback_data = callback_data;
   data->notify = notify;
   
-  error = NULL;
-
   if (first_arg_type != 0)
     {
       data->args = dbus_message_new (DBUS_MESSAGE_TYPE_METHOD_CALL);
@@ -681,10 +678,14 @@ g_daemon_file_enumerate_children (GFile      *file,
   char *obj_path;
   GDaemonFileEnumerator *enumerator;
   DBusConnection *connection;
+  char *uri;
 
   enumerator = g_daemon_file_enumerator_new ();
   obj_path = g_daemon_file_enumerator_get_object_path (enumerator);
-						       
+
+
+  uri = g_file_get_uri (file);
+  
   if (attributes == NULL)
     attributes = "";
   flags_dbus = flags;
@@ -695,8 +696,9 @@ g_daemon_file_enumerate_children (GFile      *file,
 			     DBUS_TYPE_STRING, &obj_path,
 			     DBUS_TYPE_STRING, &attributes,
 			     DBUS_TYPE_UINT32, &flags_dbus,
+			     DBUS_TYPE_STRING, &uri,
 			     0);
-  
+  g_free (uri);
   g_free (obj_path);
 
   if (reply == NULL)
@@ -718,7 +720,7 @@ g_daemon_file_enumerate_children (GFile      *file,
 static GFileInfo *
 g_daemon_file_query_info (GFile                *file,
 			  const char           *attributes,
-			  GFileQueryInfoFlags     flags,
+			  GFileQueryInfoFlags   flags,
 			  GCancellable         *cancellable,
 			  GError              **error)
 {
@@ -726,6 +728,9 @@ g_daemon_file_query_info (GFile                *file,
   dbus_uint32_t flags_dbus;
   DBusMessageIter iter;
   GFileInfo *info;
+  char *uri;
+
+  uri = g_file_get_uri (file);
 
   if (attributes == NULL)
     attributes = "";
@@ -735,8 +740,12 @@ g_daemon_file_query_info (GFile                *file,
 			     NULL, NULL,
 			     cancellable, error,
 			     DBUS_TYPE_STRING, &attributes,
-			     DBUS_TYPE_UINT32, &flags,
+			     DBUS_TYPE_UINT32, &flags_dbus,
+			     DBUS_TYPE_STRING, &uri,
 			     0);
+
+  g_free (uri);
+  
   if (reply == NULL)
     return NULL;
 
@@ -804,6 +813,9 @@ g_daemon_file_query_info_async (GFile                      *file,
 				gpointer                    user_data)
 {
   guint32 dbus_flags;
+  char *uri;
+
+  uri = g_file_get_uri (file);
 
   dbus_flags = flags;
   do_async_path_call (file,
@@ -813,7 +825,10 @@ g_daemon_file_query_info_async (GFile                      *file,
 		      query_info_async_cb, NULL, NULL,
 		      DBUS_TYPE_STRING, &attributes,
 		      DBUS_TYPE_UINT32, &dbus_flags,
+		      DBUS_TYPE_STRING, &uri,
 		      0);
+
+  g_free (uri);
 }
 
 static GFileInfo *
@@ -1515,7 +1530,12 @@ g_daemon_file_find_enclosing_mount (GFile *file,
 
   if (mount_info->user_visible)
     {
-      mount = g_daemon_mount_new (mount_info, NULL);
+      /* if we have a daemon volume monitor then return one of it's mounts */
+      mount = g_daemon_volume_monitor_find_mount_by_mount_info (mount_info);
+      if (mount == NULL)
+        {
+          mount = g_daemon_mount_new (mount_info, NULL);
+        }
       g_mount_info_unref (mount_info);
       
       if (mount)
@@ -1615,11 +1635,8 @@ g_daemon_file_delete (GFile *file,
 		      GCancellable *cancellable,
 		      GError **error)
 {
-  GDaemonFile *daemon_file;
   DBusMessage *reply;
 
-  daemon_file = G_DAEMON_FILE (file);
-  
   reply = do_sync_path_call (file, 
 			     G_VFS_DBUS_MOUNT_OP_DELETE,
 			     NULL, NULL,
@@ -1637,11 +1654,8 @@ g_daemon_file_trash (GFile *file,
 		     GCancellable *cancellable,
 		     GError **error)
 {
-  GDaemonFile *daemon_file;
   DBusMessage *reply;
 
-  daemon_file = G_DAEMON_FILE (file);
-  
   reply = do_sync_path_call (file, 
 			     G_VFS_DBUS_MOUNT_OP_TRASH,
 			     NULL, NULL,
@@ -1659,11 +1673,8 @@ g_daemon_file_make_directory (GFile *file,
 			      GCancellable *cancellable,
 			      GError **error)
 {
-  GDaemonFile *daemon_file;
   DBusMessage *reply;
 
-  daemon_file = G_DAEMON_FILE (file);
-
   reply = do_sync_path_call (file, 
 			     G_VFS_DBUS_MOUNT_OP_MAKE_DIRECTORY,
 			     NULL, NULL,
@@ -1757,13 +1768,10 @@ g_daemon_file_set_attribute (GFile *file,
 			     GCancellable *cancellable,
 			     GError **error)
 {
-  GDaemonFile *daemon_file;
   DBusMessage *message, *reply;
   DBusMessageIter iter;
   dbus_uint32_t flags_dbus;
 
-  daemon_file = G_DAEMON_FILE (file);
-
   message = create_empty_message (file, G_VFS_DBUS_MOUNT_OP_SET_ATTRIBUTE, NULL, error);
   if (!message)
     return FALSE;
@@ -1828,7 +1836,7 @@ g_daemon_file_copy (GFile                  *source,
 		    gpointer                progress_callback_data,
 		    GError                **error)
 {
-  GDaemonFile *daemon_source, *daemon_dest;
+  GDaemonFile *daemon_source;
   DBusMessage *reply;
   char *local_path;
   char *obj_path, *dbus_obj_path;
@@ -1841,7 +1849,6 @@ g_daemon_file_copy (GFile                  *source,
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_NOT_SUPPORTED, "%s", "Move not supported");
       return FALSE;
     }
-  daemon_dest = G_DAEMON_FILE (destination);
 
   if (G_IS_DAEMON_FILE (source))
     {
@@ -1920,7 +1927,6 @@ g_daemon_file_move (GFile                  *source,
 		    gpointer                progress_callback_data,
 		    GError                **error)
 {
-  GDaemonFile *daemon_source, *daemon_dest;
   DBusMessage *reply;
   char *obj_path, *dbus_obj_path;
   dbus_uint32_t flags_dbus;
@@ -1934,9 +1940,6 @@ g_daemon_file_move (GFile                  *source,
       return FALSE;
     }
   
-  daemon_source = G_DAEMON_FILE (source);
-  daemon_dest = G_DAEMON_FILE (destination);
-
   if (progress_callback)
     {
       obj_path = g_strdup_printf ("/org/gtk/vfs/callback/%p", &obj_path);
@@ -2082,7 +2085,7 @@ g_daemon_file_file_iface_init (GFileIface *iface)
   iface->get_uri = g_daemon_file_get_uri;
   iface->get_parse_name = g_daemon_file_get_parse_name;
   iface->get_parent = g_daemon_file_get_parent;
-  iface->contains_file = g_daemon_file_contains_file;
+  iface->prefix_matches = g_daemon_file_prefix_matches;
   iface->get_relative_path = g_daemon_file_get_relative_path;
   iface->resolve_relative_path = g_daemon_file_resolve_relative_path;
   iface->get_child_for_display_name = g_daemon_file_get_child_for_display_name;
