@@ -47,6 +47,14 @@
 #include "gvfsjobqueryattributes.h"
 #include "gvfsjobenumerate.h"
 #include "gvfsdaemonprotocol.h"
+#include "gvfskeyring.h"
+
+#ifdef HAVE_GCONF
+#include <gconf/gconf-client.h>
+#endif
+
+/* We load a default workgroup from gconf */
+#define PATH_GCONF_GNOME_VFS_SMB_WORKGROUP "/system/smb/workgroup"
 
 #include <libsmbclient.h>
 
@@ -68,6 +76,9 @@ struct _GVfsBackendSmb
   GMountSource *mount_source; /* Only used/set during mount */
   int mount_try;
   gboolean mount_try_again;
+	
+  gboolean password_in_keyring;
+  GPasswordSave password_save;
   
   /* Cache */
   char *cached_server_name;
@@ -77,7 +88,9 @@ struct _GVfsBackendSmb
   SMBCSRV *cached_server;
 };
 
-G_DEFINE_TYPE (GVfsBackendSmb, g_vfs_backend_smb, G_VFS_TYPE_BACKEND);
+static char *default_workgroup = NULL;
+
+G_DEFINE_TYPE (GVfsBackendSmb, g_vfs_backend_smb, G_VFS_TYPE_BACKEND)
 
 static void
 g_vfs_backend_smb_finalize (GObject *object)
@@ -168,35 +181,59 @@ auth_callback (SMBCCTX *context,
     }
   else
     {
-      GAskPasswordFlags flags = G_ASK_PASSWORD_NEED_PASSWORD;
-      char *message;
-      
-      if (backend->domain == NULL)
-	flags |= G_ASK_PASSWORD_NEED_DOMAIN;
-      if (backend->user == NULL)
-	flags |= G_ASK_PASSWORD_NEED_USERNAME;
+      gboolean in_keyring = FALSE;
 
-      /* translators: First %s is a share name, second is a server name */
-      message = g_strdup_printf (_("Password required for share %s on %s"),
-				 share_name, server_name);
-      handled = g_mount_source_ask_password (backend->mount_source,
-					     message,
-					     username_out,
-					     domain_out,
-					     flags,
-					     &abort,
-					     &ask_password,
-					     &ask_user,
-					     &ask_domain);
-      g_free (message);
-      if (!handled)
-	goto out;
+      if (!backend->password_in_keyring)
+        {
+	  in_keyring = g_vfs_keyring_lookup_password (backend->user,
+						      backend->server,
+						      backend->domain,
+						      "smb",
+						      NULL,
+						      NULL,
+						      0,
+						      &ask_user,
+						      &ask_domain,
+						      &ask_password);
+	  backend->password_in_keyring = in_keyring;
+	}
       
-      if (abort)
-	{
-	  strncpy (username_out, "ABORT", unmaxlen);
-	  strncpy (password_out, "", pwmaxlen);
-	  goto out;
+      if (!in_keyring)
+        {
+	  GAskPasswordFlags flags = G_ASK_PASSWORD_NEED_PASSWORD;
+	  char *message;
+      
+	  if (g_vfs_keyring_is_available ())
+	    flags |= G_ASK_PASSWORD_SAVING_SUPPORTED;
+	  if (backend->domain == NULL)
+	    flags |= G_ASK_PASSWORD_NEED_DOMAIN;
+	  if (backend->user == NULL)
+	    flags |= G_ASK_PASSWORD_NEED_USERNAME;
+
+	  /* translators: First %s is a share name, second is a server name */
+	  message = g_strdup_printf (_("Password required for share %s on %s"),
+				     share_name, server_name);
+	  handled = g_mount_source_ask_password (backend->mount_source,
+						 message,
+						 username_out,
+						 domain_out,
+						 flags,
+						 &abort,
+						 &ask_password,
+						 &ask_user,
+						 &ask_domain,
+						 NULL,
+						 &(backend->password_save));
+	  g_free (message);
+	  if (!handled)
+	    goto out;
+      
+	  if (abort)
+	    {
+	      strncpy (username_out, "ABORT", unmaxlen);
+	      strncpy (password_out, "", pwmaxlen);
+	      goto out;
+	    }
 	}
 
       /* Try again if this fails */
@@ -441,6 +478,11 @@ do_mount (GVfsBackend *backend,
   smb_context->callbacks.get_cached_srv_fn    = get_cached_server;
   smb_context->callbacks.remove_cached_srv_fn = remove_cached_server;
   smb_context->callbacks.purge_cached_fn      = purge_cached;
+ 
+  /* libsmbclient frees this on it's own, so make sure 
+   * to use simple system malloc */
+  if (default_workgroup != NULL)
+    smb_context->workgroup = strdup (default_workgroup);
   
   smb_context->flags = 0;
   
@@ -489,6 +531,7 @@ do_mount (GVfsBackend *backend,
 
   op_backend->mount_source = mount_source;
   op_backend->mount_try = 0;
+  op_backend->password_save = G_PASSWORD_SAVE_NEVER;
 
   do
     {
@@ -518,6 +561,16 @@ do_mount (GVfsBackend *backend,
 			_("Failed to mount Windows share"));
       return;
     }
+
+  g_vfs_keyring_save_password (op_backend->last_user,
+			       op_backend->server,
+			       op_backend->last_domain,
+			       "smb",
+			       NULL,
+			       NULL,
+			       0,
+			       op_backend->last_password,
+			       op_backend->password_save);
   
   g_vfs_job_succeeded (G_VFS_JOB (job));
 }
@@ -587,6 +640,12 @@ do_read (GVfsBackend *backend,
   GVfsBackendSmb *op_backend = G_VFS_BACKEND_SMB (backend);
   ssize_t res;
 
+  /* For some reason requests of 65536 bytes broke for me (returned 0)
+   * Maybe some smb protocol limit
+   */
+  if (bytes_requested > 65535)
+    bytes_requested = 65535;
+  
   res = op_backend->smb_context->read (op_backend->smb_context, (SMBCFILE *)handle, buffer, bytes_requested);
 
   if (res == -1)
@@ -1236,35 +1295,9 @@ set_info_from_stat (GVfsBackendSmb *backend,
 	  
 	  if (content_type)
 	    {
-	      char *mimetype_icon, *generic_mimetype_icon, *type_icon, *p;
-	      char *icon_names[3];
-	      int i;
-
-	      mimetype_icon = g_strdup (content_type);
-	      g_strdelimit (mimetype_icon, "/", '-');
-	      
-	      p = strchr (content_type, '/');
-	      if (p == NULL)
-		p = content_type + strlen (content_type);
-	      
-	      generic_mimetype_icon = g_malloc (p - content_type + strlen ("-x-generic") + 1);
-	      memcpy (generic_mimetype_icon, content_type, p - content_type);
-	      memcpy (generic_mimetype_icon + (p - content_type), "-x-generic", strlen ("-x-generic"));
-	      generic_mimetype_icon[(p - content_type) + strlen ("-x-generic")] = 0;
-	      
-	      type_icon = "text-x-generic";
-	      
-	      i = 0;
-	      icon_names[i++] = mimetype_icon;
-	      icon_names[i++] = generic_mimetype_icon;
-	      if (strcmp (generic_mimetype_icon, type_icon) != 0 &&
-		  strcmp (mimetype_icon, type_icon) != 0) 
-		icon_names[i++] = type_icon;
-	      
-	      icon = g_themed_icon_new_from_names (icon_names, i);
-	      
-	      g_free (mimetype_icon);
-	      g_free (generic_mimetype_icon);
+              icon = g_content_type_get_icon (content_type);
+              if (G_IS_THEMED_ICON (icon))
+                g_themed_icon_append_name (G_THEMED_ICON (icon), "text-x-generic");
 	    }
 	}
       
@@ -1273,12 +1306,12 @@ set_info_from_stat (GVfsBackendSmb *backend,
 	  g_file_info_set_content_type (info, content_type);
 	  g_free (content_type);
 	}
+
+      if (icon == NULL)
+	icon = g_themed_icon_new ("text-x-generic");
       
-      if (icon)
-	{
-	  g_file_info_set_icon (info, icon);
-	  g_object_unref (icon);
-	}
+      g_file_info_set_icon (info, icon);
+      g_object_unref (icon);
   }
   
   /* Don't trust n_link, uid, gid, etc returned from libsmb, its just made up.
@@ -1739,6 +1772,9 @@ g_vfs_backend_smb_class_init (GVfsBackendSmbClass *klass)
 {
   GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   GVfsBackendClass *backend_class = G_VFS_BACKEND_CLASS (klass);
+#ifdef HAVE_GCONF
+  GConfClient *gclient;
+#endif
   
   gobject_class->finalize = g_vfs_backend_smb_finalize;
 
@@ -1762,4 +1798,29 @@ g_vfs_backend_smb_class_init (GVfsBackendSmbClass *klass)
   backend_class->make_directory = do_make_directory;
   backend_class->move = do_move;
   backend_class->try_query_settable_attributes = try_query_settable_attributes;
+
+#ifdef HAVE_GCONF
+  gclient = gconf_client_get_default ();
+  if (gclient)
+    {
+      char *workgroup;
+      
+      workgroup = gconf_client_get_string (gclient, 
+					   PATH_GCONF_GNOME_VFS_SMB_WORKGROUP, NULL);
+
+      if (workgroup && workgroup[0])
+	default_workgroup = workgroup;
+      else
+	g_free (workgroup);
+      
+      g_object_unref (gclient);
+    }
+#endif
+
+}
+
+void
+g_vfs_smb_daemon_init (void)
+{
+  g_set_application_name (_("Windows Shares Filesystem Service"));
 }
